"""pyannote speaker identification command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.audio.identification.pyannote.tool import PyannoteIdentificationTool


class PyannoteCommand(Command):
    """
    Command-line interface for running pyannote identification tool.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "inference", PyannoteInferenceCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class PyannoteInferenceCommand(Command):
    """
    Command-line interface for identifying speakers from an audio track with pyannote.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--mixed_audio",
            type=Path,
            required=True,
            help="path to the input mixed audio file, such as ``/path/to/mixed-audio.wav``",
        )
        parser.add_argument(
            "--diarization",
            type=Path,
            required=True,
            help="path to the input diarization file, such as ``/path/to/diarization.rttm``",
        )
        parser.add_argument(
            "--mono_audios",
            nargs="+",
            type=Path,
            required=True,
            help="paths to the input mono audio files, such as ``/path/to/mono-audio-1.wav /path/to/mono-audio-2.wav``",
        )
        parser.add_argument(
            "--identification",
            type=Path,
            required=True,
            help="path to the output identification file, such as ``/path/to/identification.json``",
        )
        parser.add_argument(
            "--model_names",
            nargs="+",
            type=str,
            default=[
                "pyannote/embedding",
                "speechbrain/spkrec-ecapa-voxceleb",
            ],
            help="names of the embedding models",
        )
        parser.add_argument(
            "--api_token",
            type=str,
            default=None,
            help="API token for the downloading the models from HuggingFace",
        )
        parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            help="device on which to run the inference, either 'cpu' or 'cuda'",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = PyannoteIdentificationTool(
            model_names=args.model_names,
            api_token=args.api_token,
            device=args.device,
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.inference(
            mixed_audio_path=args.mixed_audio,
            diarization_path=args.diarization,
            mono_audio_paths=args.mono_audios,
            identification_path=args.identification,
        )
        del tool



"""speaker identification command-line interface."""

import argparse

from psifx.utils.command import Command, register_command
from psifx.audio.identification.pyannote.command import PyannoteCommand


class IdentificationCommand(Command):
    """
    Command-line interface for identifying speakers in audio tracks.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "pyannote", PyannoteCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()



"""pyannote speaker diarization command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.audio.diarization.pyannote.tool import PyannoteDiarizationTool


class PyannoteCommand(Command):
    """
    Command-line interface for running pyannote diarization tool.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        from psifx.audio.diarization.command import VisualizationCommand

        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "inference", PyannoteInferenceCommand)
        register_command(subparsers, "visualization", VisualizationCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class PyannoteInferenceCommand(Command):
    """
    Command-line interface for diarizing an audio track with pyannote.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--audio",
            type=Path,
            required=True,
            help="path to the input audio file, such as ``/path/to/audio.wav``",
        )
        parser.add_argument(
            "--diarization",
            type=Path,
            required=True,
            help="path to the output diarization file, such as ``/path/to/diarization.rttm``",
        )
        parser.add_argument(
            "--num_speakers",
            type=int,
            default=None,
            help="number of speaking participants, if ignored the model will try to"
            " guess it, it is advised to specify it",
        )
        parser.add_argument(
            "--model_name",
            type=str,
            default="pyannote/speaker-diarization@2.1.1",
            help="name of the diarization model used, c.f."
            " https://huggingface.co/pyannote/speaker-diarization/tree/main/reproducible_research",
        )
        parser.add_argument(
            "--api_token",
            type=str,
            default=None,
            help="API token for the downloading the models from HuggingFace",
        )
        parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            help="device on which to run the inference, either 'cpu' or 'cuda'",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = PyannoteDiarizationTool(
            model_name=args.model_name,
            api_token=args.api_token,
            device=args.device,
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.inference(
            audio_path=args.audio,
            diarization_path=args.diarization,
            num_speakers=args.num_speakers,
        )
        del tool



"""Whisper transcription tool."""

from typing import Union, Optional

from pathlib import Path

import torch
from whisper import Whisper, load_model

from psifx.audio.transcription.tool import TranscriptionTool
from psifx.io import vtt, wav


class WhisperTranscriptionTool(TranscriptionTool):
    """
    Whisper transcription and translation tool.

    :param model_name: The name of the model to use.
    :param task: Whether to "transcribe" or "translate".
    :param device: The device where the computation should be executed.
    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        model_name: str = "small",
        task: str = "transcribe",
        device: str = "cpu",
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        assert task in ["transcribe", "translate"]

        super().__init__(
            device=device,
            overwrite=overwrite,
            verbose=verbose,
        )

        self.model_name = model_name
        self.task = task
        self.model: Whisper = load_model(model_name, device=self.device)
        # Freeze the model.
        self.model.eval()
        for param in self.model.parameters():
            param.requires_grad = False

    def inference(
        self,
        audio_path: Union[str, Path],
        transcription_path: Union[str, Path],
        language: Optional[str] = None,
    ):
        """
        Whisper's backed transcription method.

        :param audio_path: Path to the audio track.
        :param transcription_path: Path to the transcription file.
        :param language: Country-code string of the spoken language.
        :return:
        """
        audio_path = Path(audio_path)
        transcription_path = Path(transcription_path)

        if self.verbose:
            print(f"audio           =   {audio_path}")
            print(f"transcription   =   {transcription_path}")

        wav.WAVReader.check(path=audio_path)
        vtt.VTTWriter.check(path=transcription_path, overwrite=self.overwrite)

        # PRE-PROCESSING
        # Nothing to do here, the model wants the path of the audio.

        # INFERENCE
        with torch.no_grad():
            segments = self.model.transcribe(
                audio=str(audio_path),
                task=self.task,
                language=language,
                verbose=self.verbose > 1,
            )["segments"]

        # POST-PROCESSING
        vtt.VTTWriter.write(
            segments=segments, path=transcription_path, overwrite=self.overwrite
        )



"""Mediapipe's skeleton graph.'"""

from mediapipe.python.solutions.holistic import (
    PoseLandmark,
    HandLandmark,
    POSE_CONNECTIONS,
    HAND_CONNECTIONS,
    FACEMESH_TESSELATION,
)
from mediapipe.python.solutions.face_mesh import FACEMESH_NUM_LANDMARKS


N_POSE_LANDMARKS = len([p.value for p in PoseLandmark])
N_FACE_LANDMARKS = FACEMESH_NUM_LANDMARKS
N_LEFT_HAND_LANDMARKS = len([p.value for p in HandLandmark])
N_RIGHT_HAND_LANDMARKS = len([p.value for p in HandLandmark])

POSE_EDGES = tuple(POSE_CONNECTIONS)
FACE_EDGES = tuple(FACEMESH_TESSELATION)
# FACE_EDGES = tuple(FACEMESH_CONTOURS)
# FACE_EDGES = tuple(FACEMESH_TESSELATION) + tuple(FACEMESH_CONTOURS)
LEFT_HAND_EDGES = tuple(HAND_CONNECTIONS)
RIGHT_HAND_EDGES = tuple(HAND_CONNECTIONS)



"""pose estimation command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.video.pose.mediapipe.command import MediaPipeCommand
from psifx.video.pose.tool import PoseEstimationTool


class PoseEstimationCommand(Command):
    """
    Command-line interface for estimating human poses from videos.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "mediapipe", MediaPipeCommand)
        register_command(subparsers, "visualization", VisualizationCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class VisualizationCommand(Command):
    """
    Command-line interface for visualizing the poses over the video.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--video",
            type=Path,
            required=True,
            help="path to the input video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--poses",
            type=Path,
            required=True,
            help="path to the input pose archive, such as ``/path/to/poses.tar.gz``",
        )
        parser.add_argument(
            "--visualization",
            type=Path,
            required=True,
            help="path to the output visualization video file, such as ``/path/to/visualization.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--confidence_threshold",
            type=float,
            default=0.0,
            help="threshold for not displaying low confidence keypoints",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = PoseEstimationTool(
            device="cpu",
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.visualization(
            video_path=args.video,
            poses_path=args.poses,
            visualization_path=args.visualization,
            confidence_threshold=args.confidence_threshold,
        )
        del tool



"""openSMILE speech processing command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.audio.speech.opensmile.tool import (
    FEATURE_SETS,
    FEATURE_LEVELS,
    OpenSmileSpeechTool,
)


class OpenSmileCommand(Command):
    """
    Command-line interface for running OpenSmile.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "inference", OpenSmileInferenceCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class OpenSmileInferenceCommand(Command):
    """
    Command-line interface for extracting non-verbal speech features from an audio track with OpenSmile.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--audio",
            type=Path,
            required=True,
            help="path to the input audio file, such as ``/path/to/audio.wav``",
        )
        parser.add_argument(
            "--diarization",
            type=Path,
            required=True,
            help="path to the input diarization file, such as ``/path/to/diarization.rttm``",
        )
        parser.add_argument(
            "--features",
            type=Path,
            required=True,
            help="path to the output feature archive, such as ``/path/to/opensmile.tar.gz``",
        )
        parser.add_argument(
            "--feature_set",
            type=str,
            default="ComParE_2016",
            help=f"available sets: {list(FEATURE_SETS.keys())}",
        )
        parser.add_argument(
            "--feature_level",
            type=str,
            default="func",
            help=f"available levels: {list(FEATURE_LEVELS.keys())}",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = OpenSmileSpeechTool(
            feature_set=args.feature_set,
            feature_level=args.feature_level,
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.inference(
            audio_path=args.audio,
            diarization_path=args.diarization,
            features_path=args.features,
        )
        del tool



"""MediaPipe pose estimation tool."""

from typing import Any, Dict, List, Tuple, Union

import json
from pathlib import Path
from tqdm import tqdm

import numpy as np
from mediapipe.python.solutions.holistic import Holistic

from psifx.video.pose.tool import PoseEstimationTool
from psifx.video.pose.mediapipe import skeleton
from psifx.io import tar, video


class MediaPipePoseEstimationTool(PoseEstimationTool):
    """
    MediaPipe pose estimation tool.

    :param model_complexity: Complexity of the model: {0, 1, 2}, higher means more FLOPs, but also more accurate results
    :param smooth: Whether to temporally smooth the inference results to reduce the jitter.
    :param device: The device where the computation should be executed.
    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        model_complexity: int = 2,
        smooth: bool = True,
        device: str = "cpu",
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device=device,
            overwrite=overwrite,
            verbose=verbose,
        )

        assert self.device == "cpu", "Only CPU support is currently available."
        self.model_complexity = model_complexity
        self.smooth = smooth

    def process_part(
        self,
        landmarks,
        size: Tuple[int, int],
        n_points: int,
    ) -> List[float]:
        """
        Processes MediaPipe output into a simple flattened list of coordinates.

        :param landmarks: MediaPipe landmarks.
        :param size: Image resolution.
        :param n_points: Expected number of points.
        :return: Processed keypoint coordinates.
        """
        h, w = size
        if landmarks is not None:
            landmarks = [[p.x, p.y, p.visibility] for p in landmarks.landmark]
            points = np.array(landmarks, dtype=np.float32)
            points[:, 0] *= w - 1
            points[:, 1] *= h - 1
        else:
            points = np.zeros((n_points, 3), dtype=np.float32)
        return points.flatten().tolist()

    def process_pose(
        self,
        results,
        size: Tuple[int, int],
    ) -> Dict[str, Any]:
        """
        Process all the parts estimated by MediaPipe, e.g. body, face, hands.

        :param results: MediaPipe output.
        :param size: Image resolution.
        :return: Processed keypoints for every part.
        """
        return {
            "pose_keypoints_2d": self.process_part(
                landmarks=results.pose_landmarks,
                size=size,
                n_points=skeleton.N_POSE_LANDMARKS,
            ),
            "face_keypoints_2d": self.process_part(
                landmarks=results.face_landmarks,
                size=size,
                n_points=skeleton.N_FACE_LANDMARKS,
            ),
            "hand_left_keypoints_2d": self.process_part(
                landmarks=results.left_hand_landmarks,
                size=size,
                n_points=skeleton.N_LEFT_HAND_LANDMARKS,
            ),
            "hand_right_keypoints_2d": self.process_part(
                landmarks=results.right_hand_landmarks,
                size=size,
                n_points=skeleton.N_LEFT_HAND_LANDMARKS,
            ),
        }

    def inference(
        self,
        video_path: Union[str, Path],
        poses_path: Union[str, Path],
    ):
        """
        Runs MediaPipe pose estimation model on a video.

        :param video_path: Path to the video file.
        :param poses_path: Path to the pose archive.
        :return:
        """
        video_path = Path(video_path)
        poses_path = Path(poses_path)

        if self.verbose:
            print(f"video   =   {video_path}")
            print(f"poses   =   {poses_path}")

        tar.TarWriter.check(path=poses_path, overwrite=self.overwrite)

        poses = {
            "edges": {
                "pose_keypoints_2d": skeleton.POSE_EDGES,
                "face_keypoints_2d": skeleton.FACE_EDGES,
                "hand_left_keypoints_2d": skeleton.LEFT_HAND_EDGES,
                "hand_right_keypoints_2d": skeleton.RIGHT_HAND_EDGES,
            }
        }

        # We have to instantiate the model for every call, because of internal states.
        # Not that it is very costly anyway.
        with (
            Holistic(
                static_image_mode=False,
                model_complexity=self.model_complexity,
                smooth_landmarks=self.smooth,
                enable_segmentation=False,
                smooth_segmentation=False,
                refine_face_landmarks=True,
            ) as model,
            video.VideoReader(path=video_path) as video_reader,
        ):
            for i, image in enumerate(
                tqdm(
                    video_reader,
                    desc="Processing",
                    disable=not self.verbose,
                )
            ):
                h, w, _ = image.shape
                results = model.process(image)
                poses[f"{i: 015d}"] = self.process_pose(
                    results=results,
                    size=(h, w),
                )

        poses = {
            f"{k}.json": json.dumps(v)
            for k, v in tqdm(
                poses.items(),
                desc="Encoding",
                disable=not self.verbose,
            )
        }
        tar.TarWriter.write(
            dictionary=poses,
            path=poses_path,
            overwrite=self.overwrite,
            verbose=self.verbose,
        )


class MediaPipePoseEstimationAndSegmentationTool(MediaPipePoseEstimationTool):
    """
    MediaPipe pose estimation and binary segmentation tool.

    :param model_complexity: Complexity of the model: {0, 1, 2}, higher means more FLOPs, but also more accurate results
    :param smooth: Whether to temporally smooth the inference results to reduce the jitter.
    :param mask_threshold: The threshold for the binarization of the segmentation mask.
    :param device: The device where the computation should be executed.
    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        model_complexity: int = 2,
        smooth: bool = True,
        mask_threshold: float = 0.1,
        device: str = "cpu",
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        assert 0.0 <= mask_threshold <= 1.0
        super().__init__(
            model_complexity=model_complexity,
            smooth=smooth,
            device=device,
            overwrite=overwrite,
            verbose=verbose,
        )
        self.mask_threshold = mask_threshold

    def process_mask(
        self,
        mask: np.ndarray,
        size: Tuple[int, int],
        threshold: float,
    ) -> np.ndarray:
        """
        Processed the floating point 2D array into a binary array given a confidence threshold.

        :param mask: Floating point mask array
        :param size: Expected image resolution
        :param threshold: Confidence threshold
        :return: Binary mask array.
        """
        h, w = size
        if mask is not None:
            mask = np.where(
                mask < threshold,
                np.array(0, dtype=np.uint8),
                np.array(255, dtype=np.uint8),
            )
            mask = np.stack((mask,) * 3, axis=-1)
        else:
            mask = np.zeros((h, w, 3), dtype=np.uint8)
        return mask

    def inference(
        self,
        video_path: Union[str, Path],
        poses_path: Union[str, Path],
        masks_path: Union[str, Path],
    ):
        """
        Runs MediaPipe pose estimation and segmentation model on a video.

        :param video_path: Path to the video file.
        :param poses_path: Path to the pose archive.
        :param masks_path: Path to the mask video file.
        :return:
        """
        video_path = Path(video_path)
        poses_path = Path(poses_path)
        masks_path = Path(masks_path)

        if self.verbose:
            print(f"video   =   {video_path}")
            print(f"poses   =   {poses_path}")
            print(f"masks   =   {masks_path}")

        tar.TarWriter.check(path=poses_path, overwrite=self.overwrite)

        poses = {
            "edges": {
                "pose_keypoints_2d": skeleton.POSE_EDGES,
                "face_keypoints_2d": skeleton.FACE_EDGES,
                "hand_left_keypoints_2d": skeleton.LEFT_HAND_EDGES,
                "hand_right_keypoints_2d": skeleton.RIGHT_HAND_EDGES,
            }
        }

        # We have to instantiate the model for every __call__, because of internal states.
        # Not that it is very costly anyway.
        with (
            Holistic(
                static_image_mode=False,
                model_complexity=self.model_complexity,
                smooth_landmarks=self.smooth,
                enable_segmentation=True,
                smooth_segmentation=self.smooth,
                refine_face_landmarks=True,
            ) as model,
            video.VideoReader(path=video_path) as video_reader,
            video.VideoWriter(
                path=masks_path,
                input_dict={"-r": video_reader.frame_rate},
                output_dict={"-c:v": "libx264", "-crf": "0", "-pix_fmt": "yuv420p"},
                overwrite=self.overwrite,
            ) as mask_writer,
        ):
            for i, image in enumerate(
                tqdm(
                    video_reader,
                    desc="Processing",
                    disable=not self.verbose,
                )
            ):
                h, w, _ = image.shape
                results = model.process(image)
                poses[f"{i: 015d}"] = self.process_pose(
                    results=results,
                    size=(h, w),
                )
                mask = self.process_mask(
                    mask=results.segmentation_mask,
                    size=(h, w),
                    threshold=self.mask_threshold,
                )
                mask_writer.write(image=mask)

        poses = {
            f"{k}.json": json.dumps(v)
            for k, v in tqdm(
                poses.items(),
                desc="Encoding",
                disable=not self.verbose,
            )
        }
        tar.TarWriter.write(
            dictionary=poses,
            path=poses_path,
            overwrite=self.overwrite,
            verbose=self.verbose,
        )



"""WAV I/O module."""

from typing import Union

from pathlib import Path


class WAVReader:
    """
    Safe WAV reader.
    """

    @staticmethod
    def check(path: Union[str, Path]):
        """
        Checks that a file has of the correct extension and exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".wav":
            raise NameError(path)
        if not path.exists():
            raise FileNotFoundError(path)


class WAVWriter:
    """
    Safe WAV writer.
    """

    @staticmethod
    def check(path: Union[str, Path], overwrite: bool = False):
        """
        Checks that a file has of the correct extension and and verifies that we can overwrite it if it exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".wav":
            raise NameError(path)
        if path.exists() and not overwrite:
            raise FileExistsError(path)



"""command utilities."""

import argparse
from typing import Type


class Command:
    """
    Base class for defining commands.

    Command instances must implement the ``setup()`` method, and they should
    implement the ``execute()`` method if they perform any functionality beyond
    defining subparsers.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        """Setup the command-line arguments for the command.

        Args:
            parser: an ``argparse.ArgumentParser`` instance
        """
        raise NotImplementedError("subclass must implement setup()")

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        """Executes the command on the given args.

        args:
            parser: the ``argparse.ArgumentParser`` instance for the command
            args: an ``argparse.Namespace`` instance containing the arguments
                for the command
        """
        raise NotImplementedError("subclass must implement execute()")


def has_subparsers(parser: argparse.ArgumentParser) -> bool:
    """
    Checks whether the parser had subparsers.

    :param parser:
    :return:
    """
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            return True

    return False


def iter_subparsers(parser: argparse.ArgumentParser):
    """
    Iterates through the subparsers.

    :param parser:
    :return:
    """
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            for subparser in action.choices.values():
                yield subparser


class RecursiveHelpAction(argparse._HelpAction):
    def __call__(self, parser: argparse.ArgumentParser, *args, **kwargs):
        self._recurse(parser)
        parser.exit()

    @staticmethod
    def _recurse(parser: argparse.ArgumentParser):
        print("", "*" * 79, parser.format_help(), sep="\n")
        for subparser in iter_subparsers(parser):
            RecursiveHelpAction._recurse(subparser)


def register_command(
    parent: argparse.ArgumentParser,
    name: str,
    command: Type[Command],
    recursive_help=True,
) -> argparse.ArgumentParser:
    """
    Registers a command to a parent subparser and returns the newly created parser.

    :param parent:
    :param name:
    :param command:
    :param recursive_help:
    :return:
    """
    parser = parent.add_parser(
        name,
        help=command.__doc__.splitlines()[0],
        description=command.__doc__.rstrip(),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.set_defaults(execute=lambda args: command.execute(parser, args))
    command.setup(parser)

    if recursive_help and has_subparsers(parser):
        parser.add_argument(
            "--all-help",
            action=RecursiveHelpAction,
            help="show help recursively and exit",
        )

    return parser


def register_main_command(
    command: Type[Command],
    version: str = None,
    recursive_help: bool = True,
) -> argparse.ArgumentParser:
    """
    Registers the main command entrypoint and returns the parser.

    :param command:
    :param version:
    :param recursive_help:
    :return:
    """
    parser = argparse.ArgumentParser(description=command.__doc__.rstrip())

    parser.set_defaults(execute=lambda args: command.execute(parser, args))
    command.setup(parser)

    if version:
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=version,
            help="show version info",
        )

    if recursive_help and has_subparsers(parser):
        parser.add_argument(
            "--all-help",
            action=RecursiveHelpAction,
            help="show help recursively and exit",
        )

    return parser



"""audio manipulation command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.audio.manipulation.tool import ManipulationTool


class ManipulationCommand(Command):
    """
    Command-line interface for manipulating audio tracks.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "extraction", ExtractionCommand)
        register_command(subparsers, "conversion", ConversionCommand)
        register_command(subparsers, "mixdown", MixDownCommand)
        register_command(subparsers, "normalization", NormalizationCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class ExtractionCommand(Command):
    """
    Command-line interface for extracting the audio track from a video.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--video",
            type=Path,
            required=True,
            help="path to the input video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--audio",
            type=Path,
            required=True,
            help="path to the output audio file, such as ``/path/to/audio.wav``",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = ManipulationTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.extraction(
            video_path=args.video,
            audio_path=args.audio,
        )
        del tool


class ConversionCommand(Command):
    """
    Command-line interface for converting any audio track to a mono audio track at 16kHz sample rate.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--audio",
            type=Path,
            required=True,
            help="path to the input audio file, such as ``/path/to/audio.wav`` (or .mp3, etc.)",
        )
        parser.add_argument(
            "--mono_audio",
            type=Path,
            required=True,
            help="path to the output audio file, such as ``/path/to/mono-audio.wav``",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = ManipulationTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.convert(
            audio_path=args.audio,
            mono_audio_path=args.mono_audio,
        )
        del tool


class MixDownCommand(Command):
    """
    Command-line interface for mixing multiple mono audio tracks.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--mono_audios",
            nargs="+",
            type=Path,
            required=True,
            help="paths to the input mono audio files, such as ``/path/to/mono-audio-1.wav /path/to/mono-audio-2.wav``",
        )
        parser.add_argument(
            "--mixed_audio",
            type=Path,
            required=True,
            help="path to the output mixed audio file, such as ``/path/to/mixed-audio.wav``",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = ManipulationTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.mixdown(
            mono_audio_paths=args.mono_audios,
            mixed_audio_path=args.mixed_audio,
        )
        del tool


class NormalizationCommand(Command):
    """
    Command-line interface for normalizing an audio track.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--audio",
            type=Path,
            required=True,
            help="path to the input audio file, such as ``/path/to/audio.wav``",
        )
        parser.add_argument(
            "--normalized_audio",
            type=Path,
            required=True,
            help="path to the output normalized audio file, such as ``/path/to/normalized-audio.wav``",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = ManipulationTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.normalization(
            audio_path=args.audio,
            normalized_audio_path=args.normalized_audio,
        )
        del tool



"""pose estimation tool."""

from typing import Union

import json
from pathlib import Path
from PIL import Image
from tqdm import tqdm

import numpy as np

from psifx.video.tool import VideoTool
from psifx.utils import draw
from psifx.io import tar, video


class PoseEstimationTool(VideoTool):
    """
    Base class for pose estimation tools from video.
    """

    def inference(
        self,
        video_path: Union[str, Path],
        poses_path: Union[str, Path],
    ):
        """
        Template of the inference method.

        :param video_path: Path to the video file.
        :param poses_path: Path to the pose archive.
        :return:
        """
        video_path = Path(video_path)
        poses_path = Path(poses_path)

        # video = load(video_path)
        # video = pre_process_func(video)
        # poses = model(video)
        # poses = post_process_func(poses)
        # poses.update({"edges": self.edges})
        # write(poses, poses_path)

        raise NotImplementedError

    def visualization(
        self,
        video_path: Union[str, Path],
        poses_path: Union[str, Path],
        visualization_path: Union[str, Path],
        confidence_threshold: float = 0.0,
    ):
        """
        Renders a visualization where the estimated poses are overlaid on top of the video.

        :param video_path: Path to the video file.
        :param poses_path: Path to the pose archive.
        :param visualization_path: Path to the visualization file.
        :param confidence_threshold: Threshold for not displaying low confidence keypoints.
        :return:
        """
        assert 0.0 <= confidence_threshold <= 1.0

        video_path = Path(video_path)
        poses_path = Path(poses_path)
        visualization_path = Path(visualization_path)

        if self.verbose:
            print(f"video           =   {video_path}")
            print(f"poses           =   {poses_path}")
            print(f"visualization   =   {visualization_path}")

        assert video_path != visualization_path
        tar.TarReader.check(path=poses_path)

        poses = tar.TarReader.read(
            poses_path,
            verbose=self.verbose,
        )

        try:
            edges = poses.pop("edges.json")
            edges = json.loads(edges)
            edges = {k: tuple(v) for k, v in edges.items()}
        except KeyError:
            print("Missing or incorrect edges.json, only the landmarks will be drawn.")
            pose = next(iter(poses.values()))
            pose = json.loads(pose)
            edges = {key: () for key, value in pose.items()}

        poses = {
            int(k.replace(".json", "")): json.loads(v)
            for k, v in tqdm(
                poses.items(),
                desc="Decoding",
                disable=not self.verbose,
            )
        }

        with (
            video.VideoReader(path=video_path) as video_reader,
            video.VideoWriter(
                path=visualization_path,
                input_dict={"-r": video_reader.frame_rate},
                output_dict={
                    "-c:v": "libx264",
                    "-crf": "15",
                    "-pix_fmt": "yuv420p",
                },
                overwrite=self.overwrite,
            ) as visualization_writer,
        ):
            for image, pose in zip(
                tqdm(
                    video_reader,
                    desc="Processing",
                    disable=not self.verbose,
                ),
                poses.values(),
            ):
                image = Image.fromarray(image.copy())
                for key, value in pose.items():
                    value = np.array(value).reshape(-1, 3)
                    points = value[..., :-1]
                    confidences = value[..., -1:]
                    image = draw.draw_pose(
                        image=image,
                        points=points,
                        edges=edges[key],
                        confidences=confidences >= confidence_threshold,
                        circle_radius=3 if "face" not in key else 0,
                        circle_thickness=1 if "face" not in key else 0,
                        line_thickness=3 if "face" not in key else 1,
                    )
                    # Single color for face? Ellipse with image relative thickness?
                image = np.asarray(image)
                visualization_writer.write(image=image)



"""``psifx`` base tool."""

from typing import Union


class Tool:
    """
    Base class for any tool.

    :param device: The device where the computation should be executed.
    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        device: str,
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        self.device = device
        self.overwrite = overwrite
        self.verbose = verbose



"""``psifx`` command-line interface."""

import argparse

import psifx
from psifx.utils.command import Command, register_command, register_main_command
from psifx.video.command import VideoCommand
from psifx.audio.command import AudioCommand


class PsifxCommand(Command):
    """``psifx`` command-line interface."""

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "audio", AudioCommand)
        register_command(subparsers, "video", VideoCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


def get_parser() -> argparse.ArgumentParser:
    """
    Create a parser for the command-line interface.
    :return:
    """
    return register_main_command(PsifxCommand, version=psifx.__version__)


def main():
    """
    Entrypoint of the psifx command-line interface.
    :return:
    """
    parser = get_parser()
    args = parser.parse_args()
    args.execute(args)


if __name__ == "__main__":
    main()



"""Whisper transcription command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.audio.transcription.whisper.tool import WhisperTranscriptionTool


class WhisperCommand(Command):
    """
    Command-line interface for running Whisper.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        from psifx.audio.transcription.command import EnhancedTranscriptionCommand

        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "inference", WhisperTranscriptionCommand)
        register_command(subparsers, "enhance", EnhancedTranscriptionCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class WhisperTranscriptionCommand(Command):
    """
    Command-line interface for transcribing an audio track with Whisper.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--audio",
            type=Path,
            required=True,
            help="path to the input audio file, such as ``/path/to/audio.wav``",
        )
        parser.add_argument(
            "--transcription",
            type=Path,
            required=True,
            help="path to the output transcription file, such as ``/path/to/transcription.vtt``",
        )
        parser.add_argument(
            "--language",
            type=str,
            default=None,
            help="language of the audio, if ignore, the model will try to guess it, "
            "it is advised to specify it",
        )
        parser.add_argument(
            "--model_name",
            type=str,
            default="small",
            help="name of the model, check "
            "https://github.com/openai/whisper#available-models-and-languages",
        )
        parser.add_argument(
            "--translate_to_english",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="whether to transcribe the audio in its original language or"
            " to translate it to english",
        )
        parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            help="device on which to run the inference, either 'cpu' or 'cuda'",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = WhisperTranscriptionTool(
            model_name=args.model_name,
            task="transcribe" if not args.translate_to_english else "translate",
            device=args.device,
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.inference(
            audio_path=args.audio,
            transcription_path=args.transcription,
            language=args.language,
        )
        del tool



"""speech processing tool."""

from typing import Union

from pathlib import Path

from psifx.tool import Tool


class SpeechTool(Tool):
    """
    Base class for non-verbal speech processing tools.
    """

    def inference(
        self,
        audio_path: Union[str, Path],
        diarization_path: Union[str, Path],
        features_path: Union[str, Path],
    ):
        """
        Template of the inference method.

        :param audio_path: Path to the audio track.
        :param diarization_path: Path to the diarization file.
        :param features_path: Path to the feature file.
        :return:
        """
        audio_path = Path(audio_path)
        diarization_path = Path(diarization_path)
        features_path = Path(features_path)

        # audio = load(audio_path)
        # audio = pre_process_func(audio)
        # features = model(audio)
        # features = post_process_func(features)
        # write(features, features_path)

        raise NotImplementedError



"""Video manipulation tool."""

from typing import Optional, Union

from pathlib import Path
import ffmpeg

from psifx.video.tool import VideoTool


class ManipulationTool(VideoTool):
    """
    Video manipulation tool.

    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device="cpu",
            overwrite=overwrite,
            verbose=verbose,
        )

    def process(
        self,
        in_video_path: Union[str, Path],
        out_video_path: Union[str, Path],
        start: Optional[float] = None,
        end: Optional[float] = None,
        x_min: Optional[int] = None,
        y_min: Optional[int] = None,
        x_max: Optional[int] = None,
        y_max: Optional[int] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ):
        """
        Offers to trim, crop and resize your video (in that exact order).

        :param in_video_path: Path to input video.
        :param out_video_path: Path to output video.
        :param start: Start timestamp to trim the video.
        :param end: End timestamp to trim the video.
        :param x_min: Top coordinate to crop the video.
        :param y_min: Left coordinate to crop the video.
        :param x_max: Bottom coordinate to crop the video.
        :param y_max: Right coordinate to crop the video.
        :param width: Width to resize the video.
        :param height: Height to resize the video.
        :return:
        """
        in_video_path = Path(in_video_path)
        out_video_path = Path(out_video_path)

        if self.verbose:
            print(f"in_video    =   {in_video_path}")
            print(f"out_video   =   {out_video_path}")

        assert in_video_path != out_video_path

        crop = all(p is not None for p in [x_min, y_min, x_max, y_max])
        no_crop = all(p is None for p in [x_min, y_min, x_max, y_max])
        resize = all(p is not None for p in [width, height])
        no_resize = all(p is None for p in [width, height])

        assert crop or no_crop
        assert resize or no_resize

        kwargs = {}
        if start is not None:
            kwargs.update(ss=start)
        if end is not None:
            kwargs.update(to=end)
        input = ffmpeg.input(str(in_video_path), **kwargs)

        video = input.video
        audio = input.audio
        if crop:
            assert x_min < x_max and y_min < y_max
            video = video.crop(
                x=x_min,
                y=y_min,
                width=x_max - x_min,
                height=y_max - y_min,
            )

        if resize:
            video = video.filter("scale", f"{width}x{height}")

        output = ffmpeg.output(video, audio, str(out_video_path))

        if out_video_path.exists():
            if self.overwrite:
                out_video_path.unlink()
            else:
                raise FileExistsError(out_video_path)
        out_video_path.parent.mkdir(parents=True, exist_ok=True)

        output.overwrite_output().run(quiet=self.verbose <= 1)



"""pyannote speaker diarization tool."""

from typing import Optional, Union

from pathlib import Path
from tqdm import tqdm

import torch
from pyannote.audio import Pipeline
from pyannote.core import Annotation

from psifx.audio.diarization.tool import DiarizationTool
from psifx.io import rttm, wav


class PyannoteDiarizationTool(DiarizationTool):
    """
    pyannote diarization tool.

    :param model_name: The name of the model to use.
    :param api_token: The HuggingFace API token to use.
    :param device: The device where the computation should be executed.
    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        model_name: str = "pyannote/speaker-diarization-3.1",
        api_token: Optional[str] = None,
        device: str = "cpu",
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device=device,
            overwrite=overwrite,
            verbose=verbose,
        )

        self.model_name = model_name
        self.api_token = api_token

        self.model: Pipeline = Pipeline.from_pretrained(
            checkpoint_path=model_name,
            use_auth_token=api_token,
        ).to(device=torch.device(device))

    def inference(
        self,
        audio_path: Union[str, Path],
        diarization_path: Union[str, Path],
        num_speakers: Optional[int] = None,
    ):
        """
        Implementation of pyannote's diarization inference method.

        :param audio_path: Path to the audio track.
        :param diarization_path: Path to the diarization file.
        :param num_speakers: Number of speaking participants, if ignored the model will try to guess it, it is advised to specify it.
        :return:
        """
        audio_path = Path(audio_path)
        diarization_path = Path(diarization_path)

        if self.verbose:
            print(f"audio           =   {audio_path}")
            print(f"diarization     =   {diarization_path}")

        wav.WAVReader.check(path=audio_path)
        rttm.RTTMWriter.check(path=diarization_path, overwrite=self.overwrite)

        # PRE-PROCESSING
        # Nothing to do here, the model wants the path of the audio.

        # INFERENCE
        for _ in tqdm(
            range(1),
            desc="Processing",
            disable=not self.verbose,
        ):
            # results is a pyannote.core.Annotation
            results: Annotation = self.model(
                file=audio_path,
                num_speakers=num_speakers,
            )

        segments = [
            {
                "type": "SPEAKER",
                "file_stem": audio_path.stem,
                "channel": 1,
                "start": segment.start,
                "duration": segment.duration,
                "orthography": "<NA>",
                "speaker_type": "<NA>",
                "speaker_name": speaker_name,
                "confidence_score": "<NA>",
                "signal_lookahead_time": "<NA>",
            }
            for segment, track_name, speaker_name in tqdm(
                results.itertracks(yield_label=True),
                desc="Parsing",
                disable=not self.verbose,
            )
        ]
        rttm.RTTMWriter.write(
            segments=segments, path=diarization_path, overwrite=self.overwrite
        )



"""speaker identification tool."""

from typing import Union, Sequence

from pathlib import Path

from psifx.tool import Tool


class IdentificationTool(Tool):
    """
    Base class for identification tools.
    """

    def inference(
        self,
        mixed_audio_path: Union[str, Path],
        diarization_path: Union[str, Path],
        mono_audio_paths: Sequence[Union[str, Path]],
        identification_path: Union[str, Path],
    ):
        """
        Template of the inference method.

        :param mixed_audio_path: Path to the audio track.
        :param diarization_path: Path to the diarization file.
        :param mono_audio_paths: Path to the mono audio tracks.
        :param identification_path: Path to the identification file.
        :return:
        """

        raise NotImplementedError



"""OpenFace face analysis tool."""

from typing import Optional, Union

import shlex
import shutil
import subprocess
from pathlib import Path
from PIL import Image
import json
import time
from tqdm import tqdm

import pandas as pd
import numpy as np
from numpy import ndarray

from psifx.video.face.tool import FaceAnalysisTool
from psifx.video.face.openface import skeleton, fields
from psifx.io import tar, video
from psifx.utils import draw

EXECUTABLE_PATH = shutil.which("FeatureExtraction")
if EXECUTABLE_PATH is not None:
    EXECUTABLE_PATH = Path(EXECUTABLE_PATH).resolve(strict=True)
DEFAULT_OPTIONS = "-2Dfp -3Dfp -pdmparams -pose -aus -gaze -au_static"


def gaze_vector_2d(
    eye_2d: ndarray,
    gaze_3d,
    depth: float,
    K: ndarray,
    K_inverse: ndarray,
):
    """
    Projects the gaze vector in 2D starting from the center of the eye.

    :param eye_2d:
    :param gaze_3d:
    :param depth:
    :param K:
    :param K_inverse:
    :return:
    """
    eye_center_2d = eye_2d[[21, 23, 25, 27]].mean(axis=-2)
    eye_center_2d = np.concatenate([eye_center_2d, np.ones(1)])
    eye_center_3d = depth * K_inverse @ eye_center_2d
    gaze_keypoint_3d = eye_center_3d + 0.1 * gaze_3d
    gaze_keypoint_2d = K @ (gaze_keypoint_3d / np.maximum(gaze_keypoint_3d[-1], 1e-8))
    return eye_center_2d[:-1], gaze_keypoint_2d[:-1]


class OpenFaceTool(FaceAnalysisTool):
    """
    OpenFace face analysis tool.

    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device="cpu",
            overwrite=overwrite,
            verbose=verbose,
        )

    def inference(
        self,
        video_path: Union[str, Path],
        features_path: Union[str, Path],
    ):
        """
        Implementation of OpenFace's face analysis inference method.

        :param video_path: The path to the video file.
        :param features_path: The path to the features archive.
        :return:
        """
        video_path = Path(video_path)
        features_path = Path(features_path)

        if self.verbose:
            print(f"video       =   {video_path}")
            print(f"features    =   {features_path}")

        assert video_path.is_file()
        tar.TarWriter.check(path=features_path, overwrite=self.overwrite)

        tmp_dir = Path(f"/tmp/TEMP_{time.time()}")
        tmp_dir.mkdir(parents=True)

        args = f"{EXECUTABLE_PATH} -f {video_path} -out_dir {tmp_dir} {DEFAULT_OPTIONS}"

        if self.verbose:
            print("OpenFace will run with the following command:")
            print(f"{args}")
            print("It might take a while, depending on the number of CPUs.")

        try:
            for i in tqdm(
                range(1),
                desc="Processing",
                disable=not self.verbose,
            ):
                process = subprocess.run(
                    args=shlex.split(args),
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                )

            if self.verbose > 1:
                print(process.stdout)

        except subprocess.CalledProcessError as error:
            print(error.stdout)

        dirty_dataframe = pd.read_csv(tmp_dir / (video_path.stem + ".csv"))

        clean_dataframe = pd.DataFrame()
        clean_dataframe["index"] = dirty_dataframe["frame"] - 1
        for clean, dirty in zip(fields.CLEAN_FIELDS, fields.DIRTY_FIELDS):
            clean_dataframe[clean] = dirty_dataframe[dirty].values.tolist()

        n_rows, _ = clean_dataframe.shape
        features = {
            "edges": {
                "eye_right_keypoints_2d": skeleton.EYE_EDGES,
                "eye_left_keypoints_2d": skeleton.EYE_EDGES,
                "eye_right_keypoints_3d": skeleton.EYE_EDGES,
                "eye_left_keypoints_3d": skeleton.EYE_EDGES,
                "face_keypoints_2d": skeleton.FACE_EDGES,
                "face_keypoints_3d": skeleton.FACE_EDGES,
            }
        }
        for i in tqdm(
            range(n_rows),
            desc="Parsing",
            disable=not self.verbose,
        ):
            index = clean_dataframe["index"][i]
            features[f"{index: 015d}"] = {
                field: np.array(clean_dataframe[field][i]).flatten().tolist()
                for field in fields.CLEAN_FIELDS
            }

        shutil.rmtree(tmp_dir)

        features = {
            f"{k}.json": json.dumps(v)
            for k, v in tqdm(
                features.items(),
                desc="Encoding",
                disable=not self.verbose,
            )
        }
        tar.TarWriter.write(
            dictionary=features,
            path=features_path,
            overwrite=self.overwrite,
            verbose=self.verbose,
        )

    def visualization(
        self,
        video_path: Union[str, Path],
        features_path: Union[str, Path],
        visualization_path: Union[str, Path],
        depth: Optional[float] = 3.0,
        f_x: Optional[float] = None,
        f_y: Optional[float] = None,
        c_x: Optional[float] = None,
        c_y: Optional[float] = None,
    ):
        """
        Produces a visualization of the face pose and eye gaze vectors over a video.

        :param video_path: The path to the video file.
        :param features_path: The path to the features archive.
        :param visualization_path: The path to the visualization file.
        :param depth: The (guesstimated) depth between the camera and the subject.
        :param f_x: The (guesstimated) focal length of the x-axis.
        :param f_y: The (guesstimated) focal length of the y-axis.
        :param c_x: The (guesstimated) principal point of the x-axis.
        :param c_y: The (guesstimated) principal point of the x-axis.
        :return:
        """
        video_path = Path(video_path)
        features_path = Path(features_path)
        visualization_path = Path(visualization_path)

        if self.verbose:
            print(f"video           =   {video_path}")
            print(f"features        =   {features_path}")
            print(f"visualization   =   {visualization_path}")

        assert video_path != visualization_path
        tar.TarReader.check(path=features_path)

        calibration = all(p is not None for p in [f_x, f_y, c_x, c_y])
        no_calibration = all(p is None for p in [f_x, f_y, c_x, c_y])
        assert calibration or no_calibration

        features = tar.TarReader.read(
            features_path,
            verbose=self.verbose,
        )

        try:
            edges = features.pop("edges.json")
            edges = json.loads(edges)
            edges = {k: tuple(v) for k, v in edges.items()}
        except KeyError:
            print("Missing or incorrect edges.json, only the landmarks will be drawn.")
            pose = next(iter(features.values()))
            pose = json.loads(pose)
            edges = {key: () for key, value in pose.items()}

        features = {
            int(k.replace(".json", "")): json.loads(v)
            for k, v in tqdm(
                features.items(),
                desc="Decoding",
                disable=not self.verbose,
            )
        }
        h, w = None, None
        K, K_inverse = None, None
        with (
            video.VideoReader(path=video_path) as video_reader,
            video.VideoWriter(
                path=visualization_path,
                input_dict={"-r": video_reader.frame_rate},
                output_dict={
                    "-c:v": "libx264",
                    "-crf": "15",
                    "-pix_fmt": "yuv420p",
                },
                overwrite=self.overwrite,
            ) as visualization_writer,
        ):
            for image, feature in zip(
                tqdm(
                    video_reader,
                    desc="Processing",
                    disable=not self.verbose,
                ),
                features.values(),
            ):
                h_, w_, _ = image.shape
                image = Image.fromarray(image.copy())
                for key in [
                    "face_keypoints_2d",
                    "eye_right_keypoints_2d",
                    "eye_left_keypoints_2d",
                ]:
                    points = np.array(feature[key]).reshape(-1, 2)
                    image = draw.draw_pose(
                        image=image,
                        points=points,
                        edges=edges[key],
                        circle_radius=0,
                        circle_thickness=0,
                        line_thickness=1,
                    )

                if h != h_ or w != w_:
                    h, w = h_, w_

                    if no_calibration:
                        f_x = 1600.0 / 1920.0 * w
                        f_y = 1600.0 / 1080.0 * h
                        c_x = w / 2
                        c_y = h / 2

                    K = np.array(
                        [
                            [f_x, 0.0, c_x],
                            [0.0, f_y, c_y],
                            [0.0, 0.0, 1.0],
                        ]
                    )
                    K_inverse = np.linalg.inv(K)

                center_right, gaze_right = gaze_vector_2d(
                    eye_2d=np.array(feature["eye_right_keypoints_2d"]).reshape(-1, 2),
                    gaze_3d=np.array(feature["gaze_right_3d"]),
                    depth=depth,
                    K=K,
                    K_inverse=K_inverse,
                )

                center_left, gaze_left = gaze_vector_2d(
                    eye_2d=np.array(feature["eye_left_keypoints_2d"]).reshape(-1, 2),
                    gaze_3d=np.array(feature["gaze_left_3d"]),
                    depth=depth,
                    K=K,
                    K_inverse=K_inverse,
                )

                image = draw.draw_pose(
                    image=image,
                    points=np.stack([center_right, gaze_right, center_left, gaze_left]),
                    edges=((0, 1), (2, 3)),
                    circle_radius=1,
                    circle_thickness=1,
                    line_thickness=1,
                )
                image = np.asarray(image)
                visualization_writer.write(image=image)



"""drawing utilities."""

from typing import Optional, Tuple

from colorsys import hls_to_rgb

import numpy as np
from PIL import Image, ImageDraw


def get_palette(
    num_colors: int,
    hue: float = 0.01,
    luminance: float = 0.6,
    saturation: float = 0.65,
) -> np.ndarray:
    """
    Returns a palette of colours between [0, 1].

    :param num_colors:
    :param hue:
    :param luminance:
    :param saturation:
    :return: [N, 3]
    """
    hues = np.linspace(0.0, 1.0, num_colors + 1)[:-1]
    hues += hue
    hues %= 1
    hues -= hues.astype(dtype=np.int64)
    hues = hues.tolist()
    palette = np.array([hls_to_rgb(hue, luminance, saturation) for hue in hues])
    palette *= 255.0
    return palette.astype(dtype=np.uint8)


def draw_points(
    image: Image.Image,
    points: np.ndarray,
    colors: Optional[np.ndarray] = None,
    radius: int = 2,
    thickness: int = 1,
) -> Image.Image:
    """
    Draws points on an image.

    :param image:
    :param points: [N, 2]
    :param colors: [N, C]
    :param radius:
    :return:
    """
    if colors is None:
        colors = get_palette(points.shape[-2])

    assert points.shape[-2] == colors.shape[-2]
    assert points.shape[-1] == 2

    draw = ImageDraw.Draw(image)

    for point, color in zip(points, colors):
        draw.ellipse(
            xy=[tuple(point - radius), tuple(point + radius)],
            outline=tuple(color),
            fill=(255, 255, 255),
            width=thickness,
        )

    return image


def draw_lines(
    image: Image.Image,
    start_points: np.ndarray,
    end_points: np.ndarray,
    colors: Optional[np.ndarray] = None,
    thickness: int = 3,
) -> Image.Image:
    """
    Draws lines on an image.

    :param image:
    :param start_points: [N, 2]
    :param end_points: [N, 2]
    :param colors: [N, C]
    :param thickness:
    :return:
    """
    if colors is None:
        colors = get_palette(start_points.shape[-2])

    assert start_points.shape[-2] == end_points.shape[-2] == colors.shape[-2]
    assert start_points.shape[-1] == end_points.shape[-1] == 2

    draw = ImageDraw.Draw(image)

    for start_point, end_point, color in zip(start_points, end_points, colors):
        draw.line(
            xy=[tuple(start_point), tuple(end_point)],
            fill=tuple(color),
            width=thickness,
        )

    return image


def draw_pose(
    image: Image.Image,
    points: np.ndarray,
    edges: Tuple[Tuple[int, int], ...],
    confidences: Optional[np.ndarray] = None,
    circle_colors: Optional[np.ndarray] = None,
    circle_radius: int = 2,
    circle_thickness: int = 1,
    line_colors: Optional[np.ndarray] = None,
    line_thickness: int = 3,
) -> Image.Image:
    """
    Draws a skeleton pose on an image.

    :param image: [H, W, C]
    :param points: [N, 2]
    :param confidences: [N, 1]
    :param edges: [M, 2]
    :param circle_colors: [N, C]
    :param circle_radius:
    :param circle_thickness:
    :param line_colors: [M, C]
    :param line_thickness:
    :return: [H, W, C]
    """
    if confidences is None:
        confidences = np.ones(points.shape[:-1] + (1,), dtype=np.bool_)

    assert points.ndim == confidences.ndim == 2
    assert points.shape[-2] == confidences.shape[-2]
    assert points.shape[-1] == 2

    if line_thickness > 0:
        start, end = list(zip(*edges))

        edge_confidences = confidences[..., start, :] & confidences[..., end, :]
        edge_confidences = np.concatenate([edge_confidences, edge_confidences], axis=-1)

        start_points = points[..., start, :].copy()
        start_points[~edge_confidences] = -1.0

        end_points = points[..., end, :].copy()
        end_points[~edge_confidences] = -1.0

        image = draw_lines(
            image=image,
            start_points=start_points,
            end_points=end_points,
            colors=line_colors,
            thickness=line_thickness,
        )

    if circle_radius > 0:
        points = points.copy()
        confidences = np.concatenate([confidences, confidences], axis=-1)
        points[~confidences] = -1.0

        image = draw_points(
            image=image,
            points=points,
            colors=circle_colors,
            radius=circle_radius,
            thickness=circle_thickness,
        )

    return image



"""RTTM I/O module."""

from typing import Dict, List, Union

from pathlib import Path
from tqdm import tqdm

import pandas as pd


COLUMN_NAMES = [
    "type",
    "file_stem",
    "channel",
    "start",
    "duration",
    "orthography",
    "speaker_type",
    "speaker_name",
    "confidence_score",
    "signal_lookahead_time",
]


class RTTMReader:
    """
    Safe RTTM reader.
    """

    @staticmethod
    def check(path: Union[str, Path]):
        """
        Checks that a file has of the correct extension and exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".rttm":
            raise NameError(path)
        if not path.exists():
            raise FileNotFoundError(path)

    @staticmethod
    def read(
        path: Union[str, Path],
        verbose: bool = True,
    ) -> List[Dict[str, Union[str, float]]]:
        """
        Reads and parse an RTTM file.

        :param path: Path to the file.
        :param verbose: Verbosity of the method.
        :return:
        """
        path = Path(path)
        RTTMReader.check(path=path)

        dataframe = pd.read_csv(
            path,
            sep=" ",
            header=None,
            names=COLUMN_NAMES,
        )
        segments = [
            {
                "type": dataframe.iloc[index]["type"],
                "file_stem": dataframe.iloc[index]["file_stem"],
                "channel": dataframe.iloc[index]["channel"],
                "start": dataframe.iloc[index]["start"],
                "duration": dataframe.iloc[index]["duration"],
                "orthography": dataframe.iloc[index]["orthography"],
                "speaker_type": dataframe.iloc[index]["speaker_type"],
                "speaker_name": dataframe.iloc[index]["speaker_name"],
                "confidence_score": dataframe.iloc[index]["confidence_score"],
                "signal_lookahead_time": dataframe.iloc[index]["signal_lookahead_time"],
            }
            for index in tqdm(
                range(dataframe.shape[0]),
                desc="Decoding",
                disable=not verbose,
            )
        ]
        return segments


class RTTMWriter:
    """
    Safe RTTM writer.
    """

    @staticmethod
    def check(path: Union[str, Path], overwrite: bool = False):
        """
        Checks that a file has of the correct extension and and verifies that we can overwrite it if it exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".rttm":
            raise NameError(path)
        if path.exists() and not overwrite:
            raise FileExistsError(path)

    @staticmethod
    def write(
        segments: List[Dict[str, Union[str, float]]],
        path: Union[str, Path],
        overwrite: bool = False,
    ):
        """
        Writes diarized audio segment information in the RTTM format.

        :param segments: Audio segment information.
        :param path: Path to the file.
        :param overwrite: Whether to overwrite, in case of an existing file.
        :return:
        """
        path = Path(path)
        RTTMWriter.check(path=path, overwrite=overwrite)

        dataframe = pd.DataFrame.from_records(segments)

        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists() and overwrite:
            path.unlink()

        dataframe.to_csv(
            path,
            sep=" ",
            na_rep="<NA>",
            columns=COLUMN_NAMES,
            header=False,
            index=False,
            float_format="%.3f",
        )



"""transcription tool."""

from typing import Union

from pathlib import Path

import pandas as pd

from psifx.tool import Tool
from psifx.io import vtt, rttm, json


class TranscriptionTool(Tool):
    """
    Base class for transcription tools.
    """

    def inference(
        self,
        audio_path: Union[str, Path],
        transcription_path: Union[str, Path],
    ):
        """
        Template of the inference method.

        :param audio_path: Path to the audio track.
        :param transcription_path: Path to the transcription file.
        :return:
        """
        audio_path = Path(audio_path)
        transcription_path = Path(transcription_path)

        # audio = load(audio_path)
        # audio = pre_process_func(audio)
        # transcription = model(audio)
        # transcription = post_process_func(transcription)
        # write(transcription, transcription_path)

        raise NotImplementedError

    def enhance(
        self,
        transcription_path: Union[str, Path],
        diarization_path: Union[str, Path],
        identification_path: Union[str, Path],
        enhanced_transcription_path: Union[str, Path],
    ):
        """
        Enhances an audio transcription by fusing the transcribed audio with inferred speaker diarization and identification.

        :param transcription_path: Path to the transcription file.
        :param diarization_path:  Path to the diarization file.
        :param identification_path:  Path to the identification file.
        :param enhanced_transcription_path:  Path to the enhanced diarization file.
        :return:
        """
        transcription_path = Path(transcription_path)
        diarization_path = Path(diarization_path)
        identification_path = Path(identification_path)
        enhanced_transcription_path = Path(enhanced_transcription_path)

        assert transcription_path != enhanced_transcription_path

        if self.verbose:
            print(f"transcription           =   {transcription_path}")
            print(f"diarization             =   {diarization_path}")
            print(f"identification          =   {identification_path}")
            print(f"enhanced_transcription  =   {enhanced_transcription_path}")

        vtt.VTTReader.check(path=transcription_path)
        rttm.RTTMReader.check(path=diarization_path)
        json.JSONReader.check(path=identification_path)
        vtt.VTTWriter.check(path=enhanced_transcription_path, overwrite=self.overwrite)

        transcription = vtt.VTTReader.read(transcription_path)
        transcription = pd.DataFrame.from_records(transcription)

        diarization = rttm.RTTMReader.read(diarization_path)
        diarization = pd.DataFrame.from_records(diarization)
        diarization["end"] = diarization["start"] + diarization["duration"]

        identification = json.JSONReader.read(identification_path)
        mapping = identification["mapping"]

        segments = []
        for transcription_index in range(len(transcription)):
            transcription_row = transcription.iloc[transcription_index]
            highest_iou_index, highest_iou = None, 0.0
            for diarization_index in range(len(diarization)):
                diarization_row = diarization.iloc[diarization_index]
                intersection_start = max(
                    transcription_row["start"], diarization_row["start"]
                )
                intersection_end = min(transcription_row["end"], diarization_row["end"])
                union_start = min(transcription_row["start"], diarization_row["start"])
                union_end = max(transcription_row["end"], diarization_row["end"])
                intersection_duration = max(0.0, intersection_end - intersection_start)
                union_duration = max(0.0, union_end - union_start)
                iou = intersection_duration / union_duration
                if iou > highest_iou:
                    highest_iou_index, highest_iou = diarization_index, iou
            matching_diarization_index = highest_iou_index
            if matching_diarization_index is not None:
                speaker_name = mapping[
                    diarization.iloc[matching_diarization_index]["speaker_name"]
                ]
            else:
                speaker_name = "NA"
            transcription.loc[transcription_index, "speaker"] = speaker_name

            segment = {
                "start": transcription.loc[transcription_index, "start"],
                "end": transcription.loc[transcription_index, "end"],
                "speaker": transcription.loc[transcription_index, "speaker"],
                "text": transcription.loc[transcription_index, "text"],
            }
            segments.append(segment)

        vtt.VTTWriter.write(
            segments=segments,
            path=enhanced_transcription_path,
            overwrite=self.overwrite,
            verbose=self.verbose,
        )



import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.video.manipulation.tool import ManipulationTool


class ManipulationCommand(Command):
    """
    Command-line interface for manipulating videos.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "process", ProcessCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class ProcessCommand(Command):
    """
    Command-line interface for processing videos.
    The trimming, cropping and resizing can be performed all at once, and in that order.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--in_video",
            type=Path,
            required=True,
            help="path to the input video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--out_video",
            type=Path,
            required=True,
            help="path to the output video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--start",
            type=float,
            default=None,
            help="trim: timestamp in seconds of the start of the selection",
        )
        parser.add_argument(
            "--end",
            type=float,
            default=None,
            help="trim: timestamp in seconds of the end of the selection",
        )
        parser.add_argument(
            "--x_min",
            type=int,
            default=None,
            help="crop: x-axis coordinate of the top-left corner in pixels",
        )
        parser.add_argument(
            "--y_min",
            type=int,
            default=None,
            help="crop: y-axis coordinate of the top-left corner in pixels",
        )
        parser.add_argument(
            "--x_max",
            type=int,
            default=None,
            help="crop: x-axis coordinate of the bottom-right corner in pixels",
        )
        parser.add_argument(
            "--y_max",
            type=int,
            default=None,
            help="crop: y-axis coordinate of the bottom-right corner in pixels",
        )
        parser.add_argument(
            "--width",
            type=int,
            default=None,
            help="resize: width of the resized output",
        )
        parser.add_argument(
            "--height",
            type=int,
            default=None,
            help="resize: height of the resized output",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = ManipulationTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.process(
            in_video_path=args.in_video,
            out_video_path=args.out_video,
            start=args.start,
            end=args.end,
            x_min=args.x_min,
            y_min=args.y_min,
            x_max=args.x_max,
            y_max=args.y_max,
            width=args.width,
            height=args.height,
        )
        del tool



"""VTT I/O module."""

from typing import Dict, Sequence, Union

import re
from pathlib import Path
from tqdm import tqdm

TIMEFRAME_LINE_PATTERN = re.compile(
    r"\s*((?:\d+:)?\d{2}:\d{2}.\d{3})\s*"
    + " --> "
    + r"\s*((?:\d+:)?\d{2}:\d{2}.\d{3})\s*"
)
SPEAKER_PATTERN = re.compile(r"<v .+>")


def seconds2timestamp(seconds: float) -> str:
    """
    Converts floating point seconds into timestamp string.

    :param seconds: Floating point seconds.
    :return: Timestamp string.
    """
    assert seconds >= 0
    time = round(seconds * 1000.0)
    hours = time // (60 * 60 * 1000)
    time -= hours * (60 * 60 * 1000)
    minutes = time // (60 * 1000)
    time -= minutes * (60 * 1000)
    seconds = time // 1000
    time -= seconds * 1000
    milliseconds = time
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}.{milliseconds:03d}"


def timestamp2seconds(timestamp: str) -> float:
    """
    Converts timestamp string into floating point seconds.

    :param timestamp: Timestamp string.
    :return: Floating point seconds.
    """
    parts = timestamp.split(":")
    assert len(parts) == 3
    sub_parts = parts[2].split(".")
    assert len(sub_parts) == 2
    hours = int(parts[0])
    minutes = int(parts[1])
    seconds = int(sub_parts[0])
    milliseconds = int(sub_parts[1])
    return hours * 60.0 * 60.0 + minutes * 60.0 + seconds + milliseconds / 1000.0


class VTTReader:
    """
    Safe VTT subtitle reader.
    """

    @staticmethod
    def check(path: Union[str, Path]):
        """
        Checks that a file has of the correct extension and exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".vtt":
            raise NameError(path)
        if not path.exists():
            raise FileNotFoundError(path)

    @staticmethod
    def read(
        path: Union[str, Path],
        verbose: bool = True,
    ) -> Sequence[Dict[str, Union[float, str]]]:
        """
        Reads and parses a VTT file.

        :param path: Path to the file.
        :param verbose: Verbosity of the method.
        :return: Subtitle segments.
        """
        path = Path(path)
        VTTReader.check(path=path)

        with path.open(encoding="utf-8") as file:
            lines = file.readlines()

        start_indexes = []
        for i, line in enumerate(lines):
            line = line.strip()
            if i == 0:
                assert line == "WEBVTT"
            match = re.match(TIMEFRAME_LINE_PATTERN, line)
            if match is not None:
                start_indexes.append(i)
        end_indexes = start_indexes[1:] + [-1]

        segments = []
        for start_index, end_index in zip(
            tqdm(
                start_indexes,
                desc="Reading",
                disable=not verbose,
            ),
            end_indexes,
        ):
            timestamps = lines[start_index].strip().split(" --> ")
            start, end = [timestamp2seconds(timestamp) for timestamp in timestamps]
            text = ""
            for line in lines[start_index + 1 : end_index]:
                text += line.strip()

            match = re.match(SPEAKER_PATTERN, text)
            if match is not None:
                speaker = match.group(0)
                text = text.replace(speaker, "")
                speaker = speaker.lstrip("<v ").rstrip(">")
            else:
                speaker = None
            segment = {
                "start": start,
                "end": end,
                "speaker": speaker,
                "text": text,
            }
            segments.append(segment)
        return segments


class VTTWriter:
    """
    Safe VTT subtitle writer.
    """

    @staticmethod
    def check(path: Union[str, Path], overwrite: bool = False):
        """
        Checks that a file has of the correct extension and and verifies that we can overwrite it if it exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".vtt":
            raise NameError(path)
        if path.exists() and not overwrite:
            raise FileExistsError(path)

    @staticmethod
    def write(
        segments: Sequence[Dict[str, Union[float, str]]],
        path: Union[str, Path],
        overwrite: bool = False,
        verbose: bool = True,
    ):
        """
        Writes transcribed audio segments into a VTT subtitle file.

        :param segments: Transcribed audio segments.
        :param path: Path to the file.
        :param overwrite: Whether to overwrite, in case of an existing file.
        :param verbose: Verbosity of the method.
        :return:
        """
        path = Path(path)
        VTTWriter.check(path=path, overwrite=overwrite)

        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists() and overwrite:
            path.unlink()

        with path.open(mode="w", encoding="utf-8") as file:
            kwargs = {"sep": "\n", "end": "\n\n", "file": file, "flush": True}

            print("WEBVTT", **kwargs)
            for segment in tqdm(
                segments,
                desc="Writing",
                disable=not verbose,
            ):
                start = seconds2timestamp(segment["start"])
                end = seconds2timestamp(segment["end"])
                text = segment["text"].strip().replace("-->", "->")
                speaker = segment.get("speaker", None)

                timeframe = f"{start} --> {end}"
                content = f"<v {speaker}>{text}" if speaker is not None else f"{text}"
                print(timeframe, content, **kwargs)



"""speaker diarization tool."""

from typing import Union

from pathlib import Path
from tqdm import tqdm

import matplotlib.pyplot as plt

from pyannote.core.annotation import Segment, Annotation
from pyannote.core import notebook

from psifx.tool import Tool
from psifx.io import rttm


class DiarizationTool(Tool):
    """
    Base class for diarization tools.
    """

    def inference(
        self,
        audio_path: Union[str, Path],
        diarization_path: Union[str, Path],
    ):
        """
        Template of the inference method.

        :param audio_path: Path to the audio track.
        :param diarization_path: Path to the diarization file.
        :return:
        """
        audio_path = Path(audio_path)
        diarization_path = Path(diarization_path)

        # audio = load(audio_path)
        # audio = pre_process_func(audio)
        # diarization = model(audio)
        # diarization = post_process_func(diarization)
        # write(diarization, diarization_path)

        raise NotImplementedError

    def visualization(
        self,
        diarization_path: Union[str, Path],
        visualization_path: Union[str, Path],
    ):
        """
        Plots a time series representing the diarized segments of an audio track.

        :param diarization_path: Path to the diarization file.
        :param visualization_path: Path to the visualization file.
        :return:
        """
        diarization_path = Path(diarization_path)
        visualization_path = Path(visualization_path)

        if self.verbose:
            print(f"diarization     =   {diarization_path}")
            print(f"visualization   =   {visualization_path}")

        rttm.RTTMReader.check(path=diarization_path)

        segments = rttm.RTTMReader.read(path=diarization_path, verbose=True)

        annotation = Annotation.from_records(
            iter(
                [
                    (
                        Segment(
                            start=segment["start"],
                            end=segment["start"] + segment["duration"],
                        ),
                        index,
                        segment["speaker_name"],
                    )
                    for index, segment in enumerate(
                        tqdm(
                            segments,
                            desc="Parsing",
                            disable=not self.verbose,
                        )
                    )
                ]
            )
        )

        if visualization_path.exists():
            if self.overwrite:
                visualization_path.unlink()
            else:
                raise FileExistsError(visualization_path)
        visualization_path.parent.mkdir(parents=True, exist_ok=True)

        plt.rcParams["figure.figsize"] = (notebook.width, 2)
        _, ax = plt.subplots()
        notebook.plot_annotation(annotation, ax=ax)
        plt.savefig(visualization_path, bbox_inches="tight")



"""OpenFace face analysis command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.video.face.openface.tool import OpenFaceTool


class OpenFaceCommand(Command):
    """
    Command-line interface for running OpenFace.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "inference", OpenFaceInferenceCommand)
        register_command(subparsers, "visualization", OpenFaceVisualizationCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class OpenFaceInferenceCommand(Command):
    """
    Command-line interface for inferring face features from videos with OpenFace.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--video",
            type=Path,
            required=True,
            help="path to the input video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--features",
            type=Path,
            required=True,
            help="path to the output feature archive, such as ``/path/to/openface.tar.gz``",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = OpenFaceTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.inference(
            video_path=args.video,
            features_path=args.features,
        )
        del tool


class OpenFaceVisualizationCommand(Command):
    """
    Command-line interface for visualizing face features from videos with OpenFace.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--video",
            type=Path,
            required=True,
            help="path to the input video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--features",
            type=Path,
            required=True,
            help="path to the input feature archive, such as ``/path/to/openface.tar.gz``",
        )
        parser.add_argument(
            "--visualization",
            type=Path,
            required=True,
            help="path to the output video file, such as ``/path/to/visualization.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--depth",
            type=float,
            default=3.0,
            help="projection: assumed static depth of the subject in meters",
        )
        parser.add_argument(
            "--f_x",
            type=float,
            default=None,
            help="projection: x-axis of the focal length",
        )
        parser.add_argument(
            "--f_y",
            type=float,
            default=None,
            help="projection: y-axis of the focal length",
        )
        parser.add_argument(
            "--c_x",
            type=float,
            default=None,
            help="projection: x-axis of the principal point",
        )
        parser.add_argument(
            "--c_y",
            type=float,
            default=None,
            help="projection: y-axis of the principal point",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = OpenFaceTool(
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.visualization(
            video_path=args.video,
            features_path=args.features,
            visualization_path=args.visualization,
            depth=args.depth,
            f_x=args.f_x,
            f_y=args.f_y,
            c_x=args.f_x,
            c_y=args.f_y,
        )
        del tool



"""audio manipulation tool."""

from typing import Sequence, Union

from pathlib import Path
import ffmpeg

from pydub import AudioSegment

from psifx.tool import Tool


class ManipulationTool(Tool):
    """
    audio manipulation tool.

    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device="cpu",
            overwrite=overwrite,
            verbose=verbose,
        )

    def extraction(
        self,
        video_path: Union[str, Path],
        audio_path: Union[str, Path],
    ):
        """
        Extracts the audio track from a video.

        :param video_path: Path to the video.
        :param audio_path: Path to the audio track.
        :return:
        """
        video_path = Path(video_path)
        audio_path = Path(audio_path)

        if self.verbose:
            print(f"video   =   {video_path}")
            print(f"audio   =   {audio_path}")

        if audio_path.exists():
            if self.overwrite:
                audio_path.unlink()
            else:
                raise FileExistsError(audio_path)
        audio_path.parent.mkdir(parents=True, exist_ok=True)
        (
            ffmpeg.input(str(video_path))
            .audio.output(str(audio_path), **{"q:a": 0, "ac": 1, "ar": 16000})
            .overwrite_output()
            .run(quiet=self.verbose <= 1)
        )

    def convert(
        self,
        audio_path: Union[str, Path],
        mono_audio_path: Union[str, Path],
    ):
        """
        Converts an audio track to a .wav audio track with 16kHz sample rate.

        :param audio_path: Path to the audio track.
        :param mono_audio_path: Path to the converted audio track.
        :return:
        """
        audio_path = Path(audio_path)
        mono_audio_path = Path(mono_audio_path)

        assert mono_audio_path.suffix == ".wav"

        if self.verbose:
            print(f"audio           =   {audio_path}")
            print(f"mono_audio      =   {mono_audio_path}")

        mono_audios = [
            audio.apply_gain(-audio.max_dBFS - 6.0)
            for audio in AudioSegment.from_file(audio_path).split_to_mono()
        ]
        mono_audio = mono_audios[0]
        for audio in mono_audios:
            mono_audio.overlay(audio)

        if mono_audio_path.exists():
            if self.overwrite:
                mono_audio_path.unlink()
            else:
                raise FileExistsError(mono_audio_path)
        mono_audio_path.parent.mkdir(parents=True, exist_ok=True)

        mono_audio.export(mono_audio_path, format="wav")

    def mixdown(
        self,
        mono_audio_paths: Sequence[Union[str, Path]],
        mixed_audio_path: Union[str, Path],
    ):
        """
        Mixes multiple mono audio tracks.

        :param mono_audio_paths: Paths to mono audio tracks.
        :param mixed_audio_path: Path to the mixed audio track.
        :return:
        """
        mono_audio_paths = [Path(path) for path in mono_audio_paths]
        mixed_audio_path = Path(mixed_audio_path)

        if self.verbose:
            print(f"mono_audios     =   {[str(path) for path in mono_audio_paths]}")
            print(f"mixed_audio     =   {mixed_audio_path}")

        mono_audios = [AudioSegment.from_wav(path) for path in mono_audio_paths]
        mono_audios = [audio.apply_gain(-audio.max_dBFS - 6.0) for audio in mono_audios]
        mixed_audio = mono_audios[0]
        for audio in mono_audios[1:]:
            mixed_audio.overlay(audio)

        if mixed_audio_path.exists():
            if self.overwrite:
                mixed_audio_path.unlink()
            else:
                raise FileExistsError(mixed_audio_path)
        mixed_audio_path.parent.mkdir(parents=True, exist_ok=True)

        mixed_audio.export(mixed_audio_path, format="wav")

    def normalization(
        self,
        audio_path: Union[str, Path],
        normalized_audio_path: Union[str, Path],
    ):
        """
        Normalizes an audio track.

        :param audio_path: Path to the audio track.
        :param normalized_audio_path: Path to the normalized audio track.
        :return:
        """
        audio_path = Path(audio_path)
        normalized_audio_path = Path(normalized_audio_path)

        if self.verbose:
            print(f"audio               =   {audio_path}")
            print(f"normalized_audio    =   {normalized_audio_path}")

        audio = AudioSegment.from_wav(audio_path)
        normalized_audio = audio.apply_gain(-audio.max_dBFS)

        if normalized_audio_path.exists():
            if self.overwrite:
                normalized_audio_path.unlink()
            else:
                raise FileExistsError(normalized_audio_path)
        normalized_audio_path.parent.mkdir(parents=True, exist_ok=True)
        normalized_audio.export(normalized_audio_path, format="wav")



"""transcription command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.audio.transcription.whisper.command import WhisperCommand
from psifx.audio.transcription.tool import TranscriptionTool


class TranscriptionCommand(Command):
    """
    Command-line interface for transcribing audio tracks.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "whisper", WhisperCommand)
        register_command(subparsers, "enhance", EnhancedTranscriptionCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class EnhancedTranscriptionCommand(Command):
    """
    Command-line interface for enhancing a transcription with diarization and identification.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--transcription",
            type=Path,
            required=True,
            help="path to the input transcription file, such as ``/path/to/transcription.vtt``",
        )
        parser.add_argument(
            "--diarization",
            type=Path,
            required=True,
            help="path to the input diarization file, such as ``/path/to/diarization.rttm``",
        )
        parser.add_argument(
            "--identification",
            type=Path,
            required=True,
            help="path to the input identification file, such as ``/path/to/identification.json``",
        )
        parser.add_argument(
            "--enhanced_transcription",
            type=Path,
            required=True,
            help="path to the output transcription file, such as ``/path/to/enhanced-transcription.vtt``",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        tool = TranscriptionTool(
            device="cpu",
            overwrite=args.overwrite,
            verbose=args.verbose,
        )
        tool.enhance(
            transcription_path=args.transcription,
            diarization_path=args.diarization,
            identification_path=args.identification,
            enhanced_transcription_path=args.enhanced_transcription,
        )
        del tool



"""pyannote speaker identification tool."""

from typing import Union, Optional, Sequence

from itertools import permutations
from pathlib import Path
from tqdm import tqdm

import numpy as np
import pandas as pd

import torch
from torch import Tensor

from pyannote.audio import Audio
from pyannote.audio.pipelines.speaker_verification import PretrainedSpeakerEmbedding
from pyannote.core import Segment

from psifx.audio.identification.tool import IdentificationTool
from psifx.io import rttm, json, wav


def cropped_waveform(
    path: Union[str, Path],
    start: float,
    end: float,
    sample_rate: int = 16000,
) -> Tensor:
    """
    Crops an audio track and returns its corresponding waveform.

    :param path: Path to the audio track.
    :param start: Start of segment in seconds.
    :param end: End of the segment in seconds.
    :param sample_rate: Sample rate of the audio track.
    :return: Tensor containing the waveform of the audio segment.
    """
    waveform, sample_rate = Audio(
        sample_rate=sample_rate,
    ).crop(
        file=path,
        segment=Segment(start, end),
        mode="pad",
    )
    return waveform


class PyannoteIdentificationTool(IdentificationTool):
    """
    pyannote speaker identification tool.

    :param model_names: The names of the models to use.
    :param api_token: The HuggingFace API token to use.
    :param device: The device where the computation should be executed.
    :param overwrite: Whether to overwrite existing files, otherwise raise an error.
    :param verbose: Whether to execute the computation verbosely.
    """

    def __init__(
        self,
        model_names: Sequence[str],
        api_token: Optional[str] = None,
        device: str = "cpu",
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device=device,
            overwrite=overwrite,
            verbose=verbose,
        )

        self.api_token = api_token
        self.models = {
            name: PretrainedSpeakerEmbedding(
                embedding=name,
                device=torch.device(device),
                use_auth_token=api_token,
            )
            for name in model_names
        }

    def inference(
        self,
        mixed_audio_path: Union[str, Path],
        diarization_path: Union[str, Path],
        mono_audio_paths: Sequence[Union[str, Path]],
        identification_path: Union[str, Path],
    ):
        """
        pyannote's backed inference method.

        :param mixed_audio_path: Path to the mixed audio track.
        :param diarization_path: Path to the diarization file.
        :param mono_audio_paths: Path to the mono audio tracks.
        :param identification_path: Path to the identification file.
        :return:
        """
        mixed_audio_path = Path(mixed_audio_path)
        diarization_path = Path(diarization_path)
        mono_audio_paths = [Path(path) for path in mono_audio_paths]
        identification_path = Path(identification_path)

        assert mixed_audio_path not in mono_audio_paths
        assert sorted(set(mono_audio_paths)) == sorted(mono_audio_paths)

        wav.WAVReader.check(path=mixed_audio_path)
        rttm.RTTMReader.check(path=diarization_path)
        for path in mono_audio_paths:
            wav.WAVReader.check(path=path)
        json.JSONWriter.check(path=identification_path, overwrite=self.overwrite)

        segments = rttm.RTTMReader.read(path=diarization_path, verbose=self.verbose)
        dataframe = pd.DataFrame.from_records(segments)
        dataframe["end"] = dataframe["start"] + dataframe["duration"]

        for name, model in tqdm(
            self.models.items(),
            desc="Processing",
            disable=not self.verbose,
        ):
            distances = []
            valids = []
            for index in tqdm(
                range(dataframe.shape[0]),
                desc="Model Embedding",
                disable=not self.verbose,
                leave=False,
            ):
                row = dataframe.iloc[index]
                if row["duration"] < 0.300:
                    distance = np.nan
                else:
                    mixed_embedding = model(
                        waveforms=cropped_waveform(
                            path=mixed_audio_path,
                            start=row["start"],
                            end=row["end"],
                        )[None, ...]
                    )
                    mono_embeddings = np.concatenate(
                        [
                            model(
                                waveforms=cropped_waveform(
                                    path=path,
                                    start=row["start"],
                                    end=row["end"],
                                )[None, ...]
                            )
                            for path in mono_audio_paths
                        ]
                    )
                    delta = mixed_embedding - mono_embeddings
                    distance = np.linalg.norm(delta, ord=2, axis=-1)

                valid = np.isfinite(distance).all()

                distances.append(distance)
                valids.append(valid)

            dataframe[f"distance_{name}"] = distances
            dataframe[f"valid_{name}"] = valids

        for name in self.models.keys():
            dataframe.drop(
                index=dataframe[~dataframe[f"valid_{name}"]].index,
                inplace=True,
            )

        mono_audio_names = [path.name for path in mono_audio_paths]

        best_mapping = None
        best_agreement = 0.0
        for speaker_names in tqdm(
            permutations(pd.Categorical(dataframe["speaker_name"]).categories.tolist()),
            desc="Voting",
            disable=not self.verbose,
        ):
            mapping = dict(zip(speaker_names, mono_audio_names))
            speaker_ids = np.stack(
                [
                    mono_audio_names.index(mapping[name])
                    for name in dataframe["speaker_name"].values
                ]
            )
            model_agreements = []
            for name in self.models.keys():
                distances = np.stack(dataframe[f"distance_{name}"].values)
                closest_ids = distances.argmin(axis=-1)
                model_agreement = (speaker_ids == closest_ids).mean()
                model_agreements.append(model_agreement)
            average_agreement = np.stack(model_agreements).mean()
            if average_agreement > best_agreement:
                best_agreement = average_agreement
                best_mapping = mapping

        data = {
            "mapping": best_mapping,
            "agreement": best_agreement,
        }
        json.JSONWriter.write(
            data=data,
            path=identification_path,
            overwrite=self.overwrite,
            verbose=self.verbose,
        )



"""Video I/O module."""

from typing import Dict, Optional, Union

from pathlib import Path

from numpy import ndarray

from skvideo.io import FFmpegReader, FFmpegWriter


class VideoReader(FFmpegReader):
    """
    Video reader object.

    :param path: The path to the video file.
    :param input_dict: Input options.
    :param output_dict: Output options.
    """

    def __init__(
        self,
        path: Union[str, Path],
        input_dict: Optional[Dict[str, str]] = None,
        output_dict: Optional[Dict[str, str]] = None,
    ):
        path = Path(path)

        assert path.exists()

        super().__init__(
            filename=str(path),
            inputdict=input_dict,
            outputdict=output_dict,
        )

        self.num_frames = self.inputframenum
        self.frame_rate = self.probeInfo["video"][self.INFO_AVERAGE_FRAMERATE]

    def __len__(self):
        return self.num_frames


class VideoWriter(FFmpegWriter):
    """
    Video writer object.

    :param path: The path to the video file.
    :param input_dict: Input options.
    :param output_dict: Output options.
    :param overwrite: Whether to overwrite existing files.
    """

    def __init__(
        self,
        path: Union[str, Path],
        input_dict: Optional[Dict[str, str]] = None,
        output_dict: Optional[Dict[str, str]] = None,
        overwrite: bool = False,
    ):
        path = Path(path)

        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists() and overwrite:
            path.unlink()

        super().__init__(
            filename=str(path),
            inputdict=input_dict,
            outputdict=output_dict,
        )

    def write(self, image: ndarray):
        """
        Appends an image to the existing video.

        :param image: [H, W, 3] ndarray.
        :return:
        """
        self.writeFrame(im=image)



"""JSON I/O module."""

from typing import Dict, List, Union

import json
from pathlib import Path
from tqdm import tqdm


class JSONReader:
    """
    Safe JSON reader.
    """

    @staticmethod
    def check(path: Union[str, Path]):
        """
        Checks that a file has of the correct extension and exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".json":
            raise NameError(path)
        if not path.exists():
            raise FileNotFoundError(path)

    @staticmethod
    def read(
        path: Union[str, Path],
        verbose: Union[bool, int] = True,
    ) -> Union[List, Dict]:
        """
        Reads and parses a JSON file.

        :param path: Path to the file.
        :param verbose: Verbosity of the method.
        :return: Deserialized data.
        """
        path = Path(path)
        JSONReader.check(path=path)

        with path.open(mode="r") as file:
            for _ in tqdm(
                range(1),
                desc="Reading",
                disable=not verbose,
            ):
                data = json.load(file)

        return data


class JSONWriter:
    """
    Safe JSON writer.
    """

    @staticmethod
    def check(path: Union[str, Path], overwrite: bool = False):
        """
        Checks that a file has of the correct extension and and verifies that we can overwrite it if it exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if path.suffix != ".json":
            raise NameError(path)
        if path.exists() and not overwrite:
            raise FileExistsError(path)

    @staticmethod
    def write(
        data: Union[List, Dict],
        path: Union[str, Path],
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        """
        Writes serializable data into a JSON file.

        :param data: Serializable data.
        :param path: Path to the file.
        :param overwrite: Whether to overwrite, in case of an existing file.
        :param verbose: Verbosity of the method.
        :return:
        """
        path = Path(path)
        JSONWriter.check(path=path, overwrite=overwrite)

        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists() and overwrite:
            path.unlink()

        with path.open(mode="w") as file:
            for _ in tqdm(range(1), desc="Writing", disable=not verbose):
                json.dump(data, file)



"""MediaPipe pose estimation command-line interface."""

import argparse
from pathlib import Path

from psifx.utils.command import Command, register_command
from psifx.video.pose.mediapipe.tool import (
    MediaPipePoseEstimationTool,
    MediaPipePoseEstimationAndSegmentationTool,
)


class MediaPipeCommand(Command):
    """
    Command-line interface for running MediaPipe.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        from psifx.video.pose.command import VisualizationCommand

        subparsers = parser.add_subparsers(title="available commands")

        register_command(subparsers, "inference", MediaPipeInferenceCommand)
        register_command(subparsers, "visualization", VisualizationCommand)

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        parser.print_help()


class MediaPipeInferenceCommand(Command):
    """
    Command-line interface for inferring human pose with MediaPipe Holistic.
    """

    @staticmethod
    def setup(parser: argparse.ArgumentParser):
        """
        Sets up the command.

        :param parser: The argument parser.
        :return:
        """
        parser.add_argument(
            "--video",
            type=Path,
            required=True,
            help="path to the input video file, such as ``/path/to/video.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--poses",
            type=Path,
            required=True,
            help="path to the output pose archive, such as ``/path/to/poses.tar.gz``",
        )
        parser.add_argument(
            "--masks",
            type=Path,
            default=None,
            help="path to the output segmentation mask video file, such as ``/path/to/masks.mp4`` (or .avi, .mkv, etc.)",
        )
        parser.add_argument(
            "--mask_threshold",
            type=float,
            default=0.1,
            help="threshold for the binarization of the segmentation mask",
        )
        parser.add_argument(
            "--model_complexity",
            type=int,
            default=2,
            help="complexity of the model: {0, 1, 2}, higher means more FLOPs, "
            "but also more accurate results",
        )
        parser.add_argument(
            "--smooth",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="temporally smooth the inference results to reduce the jitter",
        )
        parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            help="device on which to run the inference, either 'cpu' or 'cuda'",
        )
        parser.add_argument(
            "--overwrite",
            default=False,
            action=argparse.BooleanOptionalAction,
            help="overwrite existing files, otherwise raises an error",
        )
        parser.add_argument(
            "--verbose",
            default=True,
            action=argparse.BooleanOptionalAction,
            help="verbosity of the script",
        )

    @staticmethod
    def execute(parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Executes the command.

        :param parser: The argument parser.
        :param args: The arguments.
        :return:
        """
        if args.masks is None:
            tool = MediaPipePoseEstimationTool(
                model_complexity=args.model_complexity,
                smooth=args.smooth,
                device=args.device,
                overwrite=args.overwrite,
                verbose=args.verbose,
            )
            tool.inference(
                video_path=args.video,
                poses_path=args.poses,
            )
        else:
            tool = MediaPipePoseEstimationAndSegmentationTool(
                model_complexity=args.model_complexity,
                smooth=args.smooth,
                mask_threshold=args.mask_threshold,
                device=args.device,
                overwrite=args.overwrite,
                verbose=args.verbose,
            )
            tool.inference(
                video_path=args.video,
                poses_path=args.poses,
                masks_path=args.masks,
            )
        del tool



"""TAR I/O module."""

from typing import Any, Dict, Union

import tarfile
import io
from pathlib import Path
from tqdm import tqdm


class TarReader:
    """
    Safe TAR reader
    """

    @staticmethod
    def check(path: Union[str, Path]):
        """
        Checks that a file has of the correct extension and exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if ".tar" not in path.suffixes:
            raise NameError(path)
        if not path.exists():
            raise FileNotFoundError(path)

    @staticmethod
    def read(
        path: Union[str, Path],
        verbose: Union[bool, int] = True,
    ) -> Dict[str, Any]:
        """
        Extracts the content from a TAR archive.

        :param path: Path to the file.
        :param verbose: Verbosity of the method.
        :return: Extracted data.
        """
        path = Path(path)
        TarReader.check(path=path)

        with tarfile.open(path, mode="r") as tar:
            dictionary = {}
            for tarinfo in tqdm(
                tar.getmembers(),
                desc="Reading",
                disable=not verbose,
            ):
                if tarinfo.isfile():
                    key = tarinfo.name.split("/")[-1]
                    value = tar.extractfile(tarinfo).read()
                    dictionary[key] = value
        return dictionary


class TarWriter:
    """
    Safe TAR writer.
    """

    @staticmethod
    def check(path: Union[str, Path], overwrite: bool = False):
        """
        Checks that a file has of the correct extension and and verifies that we can overwrite it if it exists.

        :param path: Path to the file.
        :return:
        """
        path = Path(path)
        if ".tar" not in path.suffixes:
            raise NameError(path)
        if path.exists() and not overwrite:
            raise FileExistsError(path)

    @staticmethod
    def write(
        dictionary: Dict[str, Any],
        path: Union[str, Path],
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        """
        Write a TAR archive of the dictionary, each key/value represents a single path name and associated file.

        :param dictionary: Dictionary to be archived.
        :param path: Path to the file.
        :param overwrite: Whether to overwrite, in case of an existing file.
        :param verbose: Verbosity of the method.
        :return:
        """
        path = Path(path)
        TarWriter.check(path=path, overwrite=overwrite)

        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists() and overwrite:
            path.unlink()

        index = path.suffixes.index(".tar")
        if index == len(path.suffixes) - 1:
            compression = ""
        else:
            compression = path.suffixes[index + 1].replace(".", "")

        dir_name = path.stem.replace(".tar", "")
        with tarfile.open(path, mode=f"w:{compression}") as tar:
            for key, value in tqdm(
                dictionary.items(),
                desc="Writing",
                disable=not verbose,
            ):
                tarinfo = tarfile.TarInfo(name=f"{dir_name}/{key}")
                tarinfo.size = len(value)
                tar.addfile(tarinfo, io.BytesIO(value.encode()))



"""face analysis tool."""

from typing import Union

from pathlib import Path

from psifx.video.tool import VideoTool


class FaceAnalysisTool(VideoTool):
    """
    Base tool for face analysis.
    """

    def inference(
        self,
        video_path: Union[str, Path],
        features_path: Union[str, Path],
    ):
        """
        Template of the inference method.

        :param video_path: Path to the video file.
        :param features_path: Path to the feature archive.
        :return:
        """
        video_path = Path(video_path)
        features_path = Path(features_path)

        # video = load(video_path)
        # video = pre_process_func(video)
        # features = model(video)
        # features = post_process_func(features)
        # features.update({"metadata": metastuff})
        # write(features, features_path)

        raise NotImplementedError

    def visualization(
        self,
        video_path: Union[str, Path],
        features_path: Union[str, Path],
        visualization_path: Union[str, Path],
    ):
        """
        Template of the visualization method.

        :param video_path: Path to the video file.
        :param features_path: Path to the feature archive.
        :param visualization_path: Path to the visualization file.
        :return:
        """
        raise NotImplementedError



"""
openSMILE speech processing tool.
"""

from typing import Union

from pathlib import Path

import numpy as np
import pandas as pd

from pydub import AudioSegment
import opensmile

from psifx.audio.speech.tool import SpeechTool
from psifx.io import rttm, tar, wav


FEATURE_SETS = {
    "ComParE_2016": opensmile.FeatureSet.ComParE_2016,
    "GeMAPSv01a": opensmile.FeatureSet.GeMAPSv01a,
    "GeMAPSv01b": opensmile.FeatureSet.GeMAPSv01b,
    "eGeMAPSv01a": opensmile.FeatureSet.eGeMAPSv01a,
    "eGeMAPSv01b": opensmile.FeatureSet.eGeMAPSv01b,
    "eGeMAPSv02": opensmile.FeatureSet.eGeMAPSv02,
    "emobase": opensmile.FeatureSet.emobase,
}

FEATURE_LEVELS = {
    "lld": opensmile.FeatureLevel.LowLevelDescriptors,
    "lld_de": opensmile.FeatureLevel.LowLevelDescriptors_Deltas,
    "func": opensmile.FeatureLevel.Functionals,
}


def audio_segment_to_waveform(audio: AudioSegment) -> np.ndarray:
    """
    Converts an audio segment to a ndarray normalized waveform.

    :param audio: Audio segment
    :return: ndarray waveform.
    """
    samples = [c.get_array_of_samples() for c in audio.split_to_mono()]
    waveform = np.array(samples, dtype=np.float32) / np.iinfo(samples[0].typecode).max
    return waveform


class OpenSmileSpeechTool(SpeechTool):
    """
    openSMILE speech processing tool.
    """

    def __init__(
        self,
        feature_set: str = "ComParE_2016",
        feature_level: str = "func",
        overwrite: bool = False,
        verbose: Union[bool, int] = True,
    ):
        super().__init__(
            device="cpu",
            overwrite=overwrite,
            verbose=verbose,
        )

        self.smile = opensmile.Smile(
            feature_set=FEATURE_SETS[feature_set],
            feature_level=FEATURE_LEVELS[feature_level],
        )

    def inference(
        self,
        audio_path: Union[str, Path],
        diarization_path: Union[str, Path],
        features_path: Union[str, Path],
    ):
        """
        openSMILE inference method.

        :param audio_path: Path to the audio track.
        :param diarization_path: Path to the diarization file.
        :param features_path: Path to the feature file.
        :return:
        """
        audio_path = Path(audio_path)
        diarization_path = Path(diarization_path)
        features_path = Path(features_path)

        if self.verbose:
            print(f"audio           =   {audio_path}")
            print(f"diarization     =   {diarization_path}")
            print(f"features        =   {features_path}")

        wav.WAVReader.check(path=audio_path)
        rttm.RTTMReader.check(path=diarization_path)
        tar.TarWriter.check(path=features_path, overwrite=self.overwrite)

        audio = AudioSegment.from_wav(audio_path)

        segments = rttm.RTTMReader.read(diarization_path)

        features = []
        for segment in segments:
            start = segment["start"]
            end = segment["start"] + segment["duration"]
            start = int(start * 1000)
            end = int(end * 1000)
            cropped = audio[start:end].fade_in(duration=30).fade_out(duration=30)
            waveform = audio_segment_to_waveform(audio=cropped)
            feature = self.smile.process_signal(
                signal=waveform,
                sampling_rate=audio.frame_rate,
            )
            features.append(feature)
        features = pd.concat(features)

        tar.TarWriter.write(
            dictionary={"features.csv": features.to_csv()},
            path=features_path,
            overwrite=self.overwrite,
            verbose=self.verbose,
        )
